-- v1: name types: Tableau and Branch
-- v2: change names of functions etc.
--      - ensures that the names are appropriate, as well as
--      - fitting in with newly named types
-- v3: change to non-literate form
-- v4: some algorithmic changes: introducing HOFs etc. this is v4-1,
--     had to backtrack to accommodate mistake in introducing list
--     comprehesions.
-- v5: changes in the definintions of nub-like functions
--	- remove 'test' functions that retrun non-Bool values
--	and then test them for equality. This requires changes of
--	type.
-- v6: put nubVar into an auxiliary module
-- v7: miscellaneous changes at the end of the module, including renaming
--     foo and bar (!) and some standard function introductions. Also contra
--     renamed notContra.
-- v8: the most major refactoring
--	- changes the flow of control, so that at each iteration the
--	first `solvable' formula is attacked. Allows the per-rule
--	functions to be replaced by single function definitions;
--	- all these spearate definitions removed, as well as the
--	integer codes for `which rule to apply next';
--	- this changed flow of control means that there is now some
--	duplication in the branches: need to eliminate lists with the
--	same elements in different order - this is done by sorting all
--	branches before duplicates are removed.

import List
import ListAux	-- for nubVar

-- Defining the Data type Prop and creating the show function for it

-- This is the structure of the data type of our proposition.  It determines the ways that 
-- a Prop can be consructed, beging TRUE/FALSE (these arn't actually implomented, but can be
-- implomented if required), or a variable (eg A,B,C etc..), or type of compound formula (ie
-- AND, OR, NOT, IMPLIES and IFF).  This data type also needs to have equality tested on it

type Variable = Char
data Prop = TRUE | FALSE |
            VAR Variable |
            NOT Prop |
            AND Prop Prop |
            OR Prop Prop |
            IMPLIES Prop Prop |
            IFF Prop Prop
            deriving (Eq,Ord)

-- This is the Show function that outputs the propositions as readable sentences.  Whenever a
-- Prop is outputed, it shows it in the style created in the Show.

instance Show Prop where
    show TRUE = "TRUE"
    show FALSE = "FALSE"
    show (VAR c) = [c]
    show (NOT p) = "(" ++ show p ++ ")"
    show (AND p q) = "(" ++ show p ++ "&" ++ show q ++ ")"
    show (OR p q) = "(" ++ show p ++ "|" ++ show q ++ ")"
    show (IMPLIES p q) = "(" ++ show p ++ "=>" ++ show q ++ ")"
    show (IFF p q) = "(" ++ show p ++ "<=>" ++ show q ++ ")"

-- Addition in v1:

type Branch  = [Prop]
type Tableau = [Branch]

-- The next 3 functions are the ones that control the output to the screen of the tableau.

-- This function is the main function you use when you want to find the result.  The syntax 
-- is "Main> tableauMain [prop]" (ie "tableauMain box" will give you the tableaux result for 
-- the box problem).

tableauMain :: Tableau -> IO ()
tableauMain x = putStr (displayAllBranches (buildTableau x))

-- This function displays one branch of the tableau results, it puts every variable on a
-- new line to make it easy to look at and understand.

displayBranch :: Branch -> String
displayBranch = concat . map (++"\n") . map show

-- This function displays all the branches of the tableau, by calling the above function.
-- It then seperates each branch with a space, to make the final output clear.

displayAllBranches :: Tableau -> String
displayAllBranches = concat . intersperse "\n" . map displayBranch

-- This function is the main function that brings all the seperate parts of calculating a
-- tableaux together.

-- In v8 the branches are sorted before eliminating duplicates.

buildTableau :: Tableau -> Tableau
buildTableau = removeDuplicateBranches . map sort . removeBranchDups . removeConts . reduceRoot

-- This function finds any identical branches and removes the duplicates.  This is just here
-- to make the output simplier, because somtimes different branches give the same results as
-- others, and this just makes the output look more complex.

removeDuplicateBranches :: Tableau -> Tableau
removeDuplicateBranches = nubVar

-- This function any identical variables and removes the duplicates.  This is just here to
-- make the output simplier, by removing any unnessary duplicates of variables.

removeBranchDups :: Tableau -> Tableau
removeBranchDups = map removeBranchDup

-- This function is part of the above function, this one works on single lists of Props, 
-- where as the above one works of lists of lists of Props.

removeBranchDup :: Branch -> Branch
removeBranchDup = nubVar

-- Functions to split the Props

-- combine takes a list of Props and a list of Prop lists and adds all the values in the
-- single list of Props to all the lists in the list of Prop lists

combine :: Branch -> Tableau -> Tableau
combine ps pls = [ps ++ pl | pl<-pls ]

-- Refactored from here ... the separate functions for each
-- different rule are merged into a three ...

-- New, generic, split, remove and solve. The is function chooses
-- the first (in the list) function to `solve'.

split :: Branch -> Tableau

split ps = combine (remove ps) (solve ps)

-- Remove the first unsolved term

remove :: Branch -> Branch
remove []     = []
remove (p:ps) 
  | solved p     = p : remove ps
  | otherwise    = ps

-- This single solve function embodies all the rules in a single
-- definition.

solve :: Branch -> Tableau

solve []                      = [[]]
solve ((NOT (NOT p)):_)       = [[p]]
solve ((AND p q):_)           = [[p,q]]
solve ((OR p q):_)            = [[p],[q]]
solve ((IMPLIES p q):_)       = [[NOT p],[q]]
solve ((IFF p q):_)           = [[p,q],[NOT p,NOT q]]
solve ((NOT (AND p q)):_)     = [[NOT p],[NOT q]]
solve ((NOT (OR p q)):_)      = [[NOT p,NOT q]]
solve ((NOT (IMPLIES p q)):_) = [[p,NOT q]]
solve ((NOT (IFF p q)):_)     = [[p,NOT q],[NOT p,q]]
solve (_:ps)                  = solve ps

-- isReduced takes a list of Props and checks it for all the rules to see if it is
-- reduced to it's simplest form: i.e. it is `solved'. 
-- If it is the function returns True, if not then it returns False.

isReduced :: Branch -> Bool
isReduced = and . map solved

solved :: Prop -> Bool

solved TRUE          = True
solved FALSE         = True
solved (VAR _)       = True
solved (NOT (VAR _)) = True
solved _             = False

-- reduceRoot takes each Prop list in a list of Prop lists and checks to see if it is reduced,
-- if the Prop list is reduced the the list is returned with the result of reduceRoot on the rest
-- of the list. If the Prop list is not reduced though the list is reduced and checked for
-- contradictions, which are removed, and then passed into reduceRoot again. This has the affect
-- of recursively looping until the first Prop and any branches that may be produced from spliting
-- are delt with. Then the function returns the simplest form and processes the rest of the list
-- with reduceRoot

reduceRoot :: Tableau -> Tableau
reduceRoot [] = [[]]
reduceRoot [[]] = [[]]
reduceRoot (pl:pls) 
  = (if isReduced pl 
        then [pl] 
        else (reduceRoot (removeConts (split pl)))) 
    ++ (reduceRoot pls)

-- Remove contradictions.
-- First, tidy the resulting Tableau by removing all the empty Branches..

removeConts :: Tableau -> Tableau
removeConts = filter branchConsis . filter (/=[])

-- branchConsis recurses through the list comparing each prop
-- with every other one using 'notContra'

branchConsis :: Branch -> Bool
branchConsis []     = True
branchConsis (p:pl) = (consisAux p pl) && (branchConsis pl)
  where
  consisAux :: Prop -> Branch -> Bool
  consisAux p pl  = and [ notContra p pl2 | pl2 <- pl]

-- Return False if the 2 arguments are contradictions of each other.

notContra :: Prop -> Prop -> Bool
notContra a b = not (((NOT a) == b) || (a == (NOT b)))

-- Some example Props

plist1, plist2, plist3, plist4 :: Tableau
plist1 = [[VAR 'a', (NOT (NOT (VAR 'c'))), VAR 'b']]
plist2 = [[VAR 'a', (AND (VAR 'b') (VAR 'c')), (OR (VAR 'b') (VAR 'c'))]]
plist3 = [[(IMPLIES (VAR 'a') (VAR 'b')), (IFF (VAR 'c') (VAR 'a')), (NOT (AND (VAR 'a') (VAR 'b')))]]
plist4 = [[(NOT (OR (VAR 'a') (VAR 'b'))), (NOT (IMPLIES (VAR 'c') (VAR 'd'))), (NOT (IFF (VAR 'a') (VAR 'b')))]]

-- The questions converted into Prop form

q1, q2, q3, q4 :: Tableau
q1 = [[(IFF (NOT (OR (VAR 'a') (VAR 'b'))) (NOT (AND (NOT (VAR 'a')) (NOT (VAR 'b')))))]]
q2 = [[(NOT (IFF (OR (VAR 'a') (VAR 'b')) (NOT (AND (NOT (VAR 'a')) (NOT (VAR 'b'))))))]]
q3 = [[(NOT (IMPLIES (IMPLIES (IMPLIES (VAR 'c') (VAR 'b')) (VAR 'c')) (VAR 'c')))]]
q4 = [[(IMPLIES (IMPLIES (VAR 'a') (VAR 'b')) (IMPLIES (VAR 'b') (VAR 'a')))]]

b1, b2, b3, b4, b5, b6, b7, b8, b9 :: Prop
b1 = (IFF (VAR 'f') (NOT (VAR 'c')))
b2 = (IFF (VAR 'g') (NOT (VAR 'c')))
b3 = (IFF (VAR 'h') (VAR 'd'))
b4 = (OR (OR (VAR 'b') (VAR 'c')) (VAR 'd'))
b5 = (NOT (AND (VAR 'b') (VAR 'c')))
b6 = (NOT (AND (VAR 'c') (VAR 'd')))
b7 = (NOT (AND (VAR 'd') (VAR 'b')))
b8 = (OR (OR (VAR 'f') (VAR 'g')) (VAR 'h'))
b9 = (OR (OR (NOT (VAR 'f')) (NOT (VAR 'g'))) (NOT (VAR 'h')))

-- box and box1 are virtualy identical, except that they AND the values in a slightly 
-- different way.  (box1 was only created to work out if another function was working
-- correctly or not).

box, box1 :: Tableau
box = [[(AND (AND (AND (b1) (b2)) (AND (b3) (b4))) (AND (AND (b5) (b6)) (AND (b7) (AND (b8) (b9)))))]]
box1 = [[(AND (AND (AND (AND (b1) (b2)) (AND (b3) (b4))) (AND (AND (b5) (b6)) (AND (b7) (b8)))) (b9))]]

-- The box example as a test.

test = tableauMain box
