-- v1: name types: Tableau and Branch
-- v2: change names of functions etc.
--      - ensures that the names are appropriate, as well as
--      - fitting in with newly named types
-- v3: change to non-literate form
-- v4: some algorithmic changes: introducing HOFs etc. this is v4-1,
--     had to backtrack to accommodate mistake in introducing list
--     comprehesions.
-- v5: changes in the definintions of nub-like functions
--	- remove 'test' functions that retrun non-Bool values
--	and then test them for equality. This requires changes of
--	type.
-- v6: put nubVar into an auxiliary module
-- v7: miscellaneous changes at the end of the module, including renaming
--     foo and bar (!) and some standard function introductions. Also contra
--     renamed notContra.
-- v8: the most major refactoring
--	- changes the flow of control, so that at each iteration the
--	first `solvable' formula is attacked. Allows the per-rule
--	functions to be replaced by single function definitions;
--	- all these spearate definitions removed, as well as the
--	integer codes for `which rule to apply next';
--	- this changed flow of control means that there is now some
--	duplication in the branches: need to eliminate lists with the
--	same elements in different order - this is done by sorting all
--	branches before duplicates are removed.
-- v9: Changing Branch and Tableau into Set Prop and Set (Set Prop): uses
--	the Set library. Not too difficult, though impossible to test
--	until the whole change is made. What *can* be done is type 
--	checking (with dummy defintions if necessary) ... this is really
--	very useful, I guess in contrast with Smalltalk.

import List hiding (union)
import ListAux	-- for nubVar
import Set

-- Defining the Data type Prop and creating the show function for it

-- This is the structure of the data type of our proposition.  It determines the ways that 
-- a Prop can be consructed, beging TRUE/FALSE (these arn't actually implomented, but can be
-- implomented if required), or a variable (eg A,B,C etc..), or type of compound formula (ie
-- AND, OR, NOT, IMPLIES and IFF).  This data type also needs to have equality tested on it

type Variable = Char
data Prop = TRUE | FALSE |
            VAR Variable |
            NOT Prop |
            AND Prop Prop |
            OR Prop Prop |
            IMPLIES Prop Prop |
            IFF Prop Prop
            deriving (Eq,Ord)

-- This is the Show function that outputs the propositions as readable sentences.  Whenever a
-- Prop is outputed, it shows it in the style created in the Show.

instance Show Prop where
    show TRUE = "TRUE"
    show FALSE = "FALSE"
    show (VAR c) = [c]
    show (NOT p) = "(" ++ show p ++ ")"
    show (AND p q) = "(" ++ show p ++ "&" ++ show q ++ ")"
    show (OR p q) = "(" ++ show p ++ "|" ++ show q ++ ")"
    show (IMPLIES p q) = "(" ++ show p ++ "=>" ++ show q ++ ")"
    show (IFF p q) = "(" ++ show p ++ "<=>" ++ show q ++ ")"

-- Addition in v1:

type Branch  = Set Prop
type Tableau = Set Branch

-- The next 3 functions are the ones that control the output to the screen of the tableau.

-- This function is the main function you use when you want to find the result.  The syntax 
-- is "Main> tableauMain [prop]" (ie "tableauMain box" will give you the tableaux result for 
-- the box problem).

tableauMain :: Tableau -> IO ()
tableauMain x = putStr (displayAllBranches (buildTableau x))

-- This function displays one branch of the tableau results, it puts every variable on a
-- new line to make it easy to look at and understand.

displayBranch :: Branch -> String
displayBranch = foldSet (++) [] . mapSet (++"\n") . mapSet show

-- This function displays all the branches of the tableau, by calling the above function.
-- It then seperates each branch with a space, to make the final output clear.

displayAllBranches :: Tableau -> String
displayAllBranches = foldSet (++) [] . mapSet (++"\n") . mapSet displayBranch

-- This function is the main function that brings all the seperate parts of calculating a
-- tableaux together.

-- In v9, no need to sort or to remove duplicates: working over Sets. 
-- Can simplify the definition of buildTableau and remove the 
-- definitions no longer used. Cool.

buildTableau :: Tableau -> Tableau
buildTableau = removeConts . reduceRoot

-- Functions to split the Props

-- combine takes a list of Props and a list of Prop lists and adds all the values in the
-- single list of Props to all the lists in the list of Prop lists

combine :: Branch -> Tableau -> Tableau
combine ps pls = makeSet [ ps `union` pl | pl <- flatten pls ]



-- New, generic, split, remove and solve. The is function chooses
-- the first (in the list) function to `solve'.

split :: Branch -> Tableau

split ps = combine (remove ps) (solve firstUns)
	   where firstUns = pick (filterSet (not.solved) ps)

-- Remove the first unsolved term: 
-- NB: assumes that there is such a term.

remove :: Branch -> Branch
remove branch
  | branch==empty	= empty
  | otherwise		= branch `diff` sing firstUns
    where firstUns = pick (filterSet (not.solved) branch)

-- This single solve function embodies all the rules in a single
-- definition.

-- Modified the type of solve to take a single formula, rather than
-- a singleton set: allows the pattern match to persist in a 
-- position where it is really needed.

solve :: Prop -> Tableau

solve (NOT (NOT p))       = sing (sing p)
solve (AND p q)           = sing (makeSet [p,q])
solve (OR p q)            = makeSet [sing p, sing q]
solve (IMPLIES p q)       = makeSet [sing (NOT p),sing q]
solve (IFF p q)           = makeSet [makeSet [p,q],makeSet [NOT p,NOT q]]
solve (NOT (AND p q))     = makeSet [makeSet [NOT p],makeSet [NOT q]]
solve (NOT (OR p q))      = makeSet [makeSet [NOT p,NOT q]]
solve (NOT (IMPLIES p q)) = makeSet [makeSet [p,NOT q]]
solve (NOT (IFF p q))     = makeSet [makeSet [p,NOT q],makeSet [NOT p,q]]
solve _                   = sing empty

-- isReduced takes a list of Props and checks it for all the rules to see if it is
-- reduced to it's simplest form: i.e. it is `solved'. 
-- If it is the function returns True, if not then it returns False.

isReduced :: Branch -> Bool
isReduced = foldSet (&&) True . mapSet solved

solved :: Prop -> Bool

solved TRUE          = True
solved FALSE         = True
solved (VAR _)       = True
solved (NOT (VAR _)) = True
solved _             = False

-- reduceRoot takes each Prop list in a list of Prop lists and checks to see if it is reduced,
-- if the Prop list is reduced the the list is returned with the result of reduceRoot on the rest
-- of the list. If the Prop list is not reduced though the list is reduced and checked for
-- contradictions, which are removed, and then passed into reduceRoot again. This has the affect
-- of recursively looping until the first Prop and any branches that may be produced from spliting
-- are delt with. Then the function returns the simplest form and processes the rest of the list
-- with reduceRoot

reduceRoot :: Tableau -> Tableau
reduceRoot tab
  | tab == empty        = sing empty
  | tab == sing empty   = sing empty
  | otherwise          
      = (if isReduced pl 
        then sing pl 
        else (reduceRoot (removeConts (split pl)))) 
        `union` (reduceRoot pls)
    where pl  = pick tab
          pls = tab `diff` sing pl
          

-- Remove contradictions.
-- First, tidy the resulting Tableau by removing all the empty Branches..

removeConts :: Tableau -> Tableau
removeConts = filterSet branchConsis . filterSet (/=empty)

-- branchConsis recurses through the list comparing each prop
-- with every other one using 'notContra'


branchConsis :: Branch -> Bool
branchConsis = primRecSet aux True
  where
  aux p pl v = (consisAux p pl) && v
  consisAux :: Prop -> Branch -> Bool
  consisAux p pl  = foldSet (&&) True (mapSet (\pl2 -> notContra p pl2) pl)

-- Return False if the 2 arguments are contradictions of each other.

notContra :: Prop -> Prop -> Bool
notContra a b = not (((NOT a) == b) || (a == (NOT b)))

-- Some example Props

plist1, plist2, plist3, plist4 :: Tableau
plist1 = makeSet [makeSet [VAR 'a', (NOT (NOT (VAR 'c'))), VAR 'b']]
plist2 = makeSet [makeSet [VAR 'a', (AND (VAR 'b') (VAR 'c')), (OR (VAR 'b') (VAR 'c'))]]
plist3 = makeSet [makeSet [(IMPLIES (VAR 'a') (VAR 'b')), (IFF (VAR 'c') (VAR 'a')), (NOT (AND (VAR 'a') (VAR 'b')))]]
plist4 = makeSet [makeSet [(NOT (OR (VAR 'a') (VAR 'b'))), (NOT (IMPLIES (VAR 'c') (VAR 'd'))), (NOT (IFF (VAR 'a') (VAR 'b')))]]

-- The questions converted into Prop form

q1, q2, q3, q4 :: Tableau
q1 = makeSet [makeSet [(IFF (NOT (OR (VAR 'a') (VAR 'b'))) (NOT (AND (NOT (VAR 'a')) (NOT (VAR 'b')))))]]
q2 = makeSet [makeSet [(NOT (IFF (OR (VAR 'a') (VAR 'b')) (NOT (AND (NOT (VAR 'a')) (NOT (VAR 'b'))))))]]
q3 = makeSet [makeSet [(NOT (IMPLIES (IMPLIES (IMPLIES (VAR 'c') (VAR 'b')) (VAR 'c')) (VAR 'c')))]]
q4 = makeSet [makeSet [(IMPLIES (IMPLIES (VAR 'a') (VAR 'b')) (IMPLIES (VAR 'b') (VAR 'a')))]]

b1, b2, b3, b4, b5, b6, b7, b8, b9 :: Prop
b1 = (IFF (VAR 'f') (NOT (VAR 'c')))
b2 = (IFF (VAR 'g') (NOT (VAR 'c')))
b3 = (IFF (VAR 'h') (VAR 'd'))
b4 = (OR (OR (VAR 'b') (VAR 'c')) (VAR 'd'))
b5 = (NOT (AND (VAR 'b') (VAR 'c')))
b6 = (NOT (AND (VAR 'c') (VAR 'd')))
b7 = (NOT (AND (VAR 'd') (VAR 'b')))
b8 = (OR (OR (VAR 'f') (VAR 'g')) (VAR 'h'))
b9 = (OR (OR (NOT (VAR 'f')) (NOT (VAR 'g'))) (NOT (VAR 'h')))

-- box and box1 are virtualy identical, except that they AND the values in a slightly 
-- different way.  (box1 was only created to work out if another function was working
-- correctly or not).

box, box1 :: Tableau
box = makeSet [makeSet [(AND (AND (AND (b1) (b2)) (AND (b3) (b4))) (AND (AND (b5) (b6)) (AND (b7) (AND (b8) (b9)))))]]
box1 = makeSet [makeSet [(AND (AND (AND (AND (b1) (b2)) (AND (b3) (b4))) (AND (AND (b5) (b6)) (AND (b7) (b8)))) (b9))]]

-- The box example as a test.

test = tableauMain box


